--[[
==============================================================================
Trade Assistant 2.1 - Plugin for Vendetta Online (all platforms)
==============================================================================
Created by Mad Miner Moda in 2014, fixed and improved by Snib in 2022.

Store sellable item data relevant for trading (prices, cu and kg) and provide
a GUI for finding profitable trade routes and where item are sold.

Visit each single station in game to record their inventory. Note that some 
stations are fully dynamic (currently only Latos stations) and their inventory 
changes depending on what gets sold there.

Sell prices can only be obtained for items you have in your inventory (in your 
ship's cargohold or stored at the station you are in).

Sell prices depend on player supply, whereas buy prices are usually static. 
Commodity prices are, however, modified by the player's trade skill level and 
standing with the faction operating the station.

Commands:
   /ta              -- show GUI
   /ta fullscreen   -- toggles between fullscreen and windowed mode
   /ta target_print -- toggles printing of sell prices of current target or 
                       target's scanned cargo - prices are color coded based on 
                       credits/cu to decide at a glance what is worth taking
   /ta version      -- print Trade Assistant version
   /ta reset        -- reset database, all existing data will be lost
==============================================================================
]]

-- set to false if ship protoype mass bug ever gets fixed, see
-- https://www.vendetta-online.com/x/msgboard/2/38113
local shipMassBugWorkaround = true
local activeShipMass = 0

local plugin_path = "plugins/TradeAssist/"
local notesid = 20100201
local command = "ta"

local floor = math.floor
local fmod = math.fmod
local sub = string.sub
local format = string.format
local sorted = {}
local timer = Timer()
local init_done = false
local fullscreen = false
local is_premium = false
local is_mobile = (Platform == "Android" or Platform == "iOS")
local has_old_goods = false
local last_scanned = ""

local clicked = {}
local click_timer = Timer()
local click_timeout = function() clicked = {} end

local TA
TA = {
    stations = {}, -- list of stations and related info here.
    goods = {}, -- trade goods table. this is where the magic happens.
    goods_old = {}, -- old format trade goods table gets imported to this
    Version = "2.1~Snib",
    GUI = {
        Color = { -- also controls the sort order for the goods tab
            WHITE        = rgbtohex("255 255 255"),
            SPRINGGREEN  = rgbtohex("0 255 127"),
            GREEN        = rgbtohex("51 204 51"),
            YELLOW       = rgbtohex("255 255 0"),
            ORANGE       = rgbtohex("255 165 0"),
            RED          = rgbtohex("255 0 0"),
            commodities  = rgbtohex("55 210 255"),
            ship         = rgbtohex("126 255 128"),
            ore          = rgbtohex("180 180 180"),
            mining       = rgbtohex("210 115 185"),
            lightweapon  = rgbtohex("218 110 75"),
            heavyweapon  = rgbtohex("220 40 80"),
            turretweapon = rgbtohex("254 69 0"),
            battery      = rgbtohex("255 185 53"),
        },
        Factioncolors = {
            Gray = rgbtohex(FactionColor_RGB[0]),
            Itani = rgbtohex(FactionColor_RGB[1]),
            Serco = rgbtohex(FactionColor_RGB[2]),
            UIT = rgbtohex(FactionColor_RGB[3]),
            Developers = rgbtohex("255 0 255")
        },
        Systemownership = {
            "Serco", -- 1: Sol II
            "Serco", -- 2: Betheshee
            "Serco", -- 3: Geira Rutilus
            "Itani", -- 4: Deneb
            "Itani", -- 5: Eo
            "Itani", -- 6: Cantus
            "Itani", -- 7: Metana
            "Itani", -- 8: Setalli Shinas
            "Itani", -- 9: Itan
            "Itani", -- 10: Pherona
            "Itani", -- 11: Artana Aquilus
            "Itani", -- 12: Divinia
            "Itani", -- 13: Jallik
            "Gray", -- 14: Edras
            "UIT", -- 15: Verasi
            "Gray", -- 16: Pelatus
            "Gray", -- 17: Bractus
            "UIT", -- 18: Nyrius
            "UIT", -- 19: Dau
            "Gray", -- 20: Sedina
            "UIT", -- 21: Azek
            "Gray", -- 22: Odia
            "Gray", -- 23: Latos
            "UIT", -- 24: Arta Caelestis
            "Gray", -- 25: Ukari
            "Gray", -- 26: Helios
            "Serco", -- 27: Initros
            "Serco", -- 28: Pyronis
            "Serco", -- 29: Rhamus
            "Serco", -- 30: Dantia
            "Developers", -- 31: Devlopia
            "Developers", -- 32: (Qualcomm Demo system)
            "Gray" -- 33: Unknown System
        },
    },
    Print = function(str)
        print(TA.print_prefix .. tostring(str))
    end,
    target_print = false,
}
TA.Title = "Trade Assistant v" .. TA.Version
TA.print_prefix = TA.GUI.Color.WHITE .. "[" .. TA.GUI.Color.ORANGE .. "TA" ..
    TA.GUI.Color.WHITE .. "] "

local function save_data_now()
    timer:Kill()
    SaveSystemNotes({
        version = TA.Version,
        goods = TA.goods,
        stations = TA.stations,
        goods_old = TA.goods_old,
        target_print = TA.target_print,
        fullscreen = fullscreen,
    }, notesid)
end

local function save_data()
    timer:SetTimeout(90000, save_data_now)
end

local function toggle_target_print(state)
    if state == nil then
        TA.target_print = not TA.target_print
    else
        TA.target_print = state
    end
    if TA.target_print then
        if state == nil then
            TA.Print("Target price printing enabled")
        end
        if init_done then
            RegisterEvent(TA.EventHandler, "TARGET_CHANGED")
            RegisterEvent(TA.EventHandler, "TARGET_SCANNED")
        end
    else
        if state == nil then
            TA.Print("Target price printing disabled")
        end
        UnregisterEvent(TA.EventHandler, "TARGET_CHANGED")
        UnregisterEvent(TA.EventHandler, "TARGET_SCANNED")
    end
end

local function create_matrix(columndefs, defsort)
    local matrix
    local sortedlist
    local sort_key = defsort -- default to sort by distance
    local numcolumns = (#columndefs)
    local function set_sort_mode(mode)
        -- clicked on title of column
        sort_key = mode
        -- color the text accordingly
        for i = 1, numcolumns do
            matrix:setattribute("bgcolor", 0, i, mode == i and
                tabseltextcolor or tabunseltextcolor)
        end
    end

    local function sort_list(key, list)
        table.sort(list, columndefs[key].fn)
    end

    local function reload_matrix(list)
        matrix.numlin = (#list)

        for index, v in ipairs(list) do
            columndefs.update_entry(matrix, index, v)
        end
    end

    local update_matrix = function()
        set_sort_mode(sort_key)
        sort_list(sort_key, sortedlist)
        reload_matrix(sortedlist)
    end

    matrix = iup.pdasubsubmatrix {
        numcol = numcolumns,
        expand = "YES",
        click_cb = function(_, row, col)
            if row == 0 and tonumber(matrix.NUMLIN) > 0 then
                set_sort_mode(col)
                update_matrix()
            elseif columndefs.on_sel and sortedlist then
                columndefs.on_sel(matrix, row, sortedlist[row])
                update_matrix()
            end
        end,
        edition_cb = function()
            return iup.IGNORE
        end,
        bgcolor_cb = function(_, row)
            local c = ListColors.Numbers[fmod(row, 2)]
            return c[1], c[2], c[3], c[4], iup.DEFAULT
        end,
        update = update_matrix,
        reset = function(_, characterlist)
            matrix.dellin = "1--1"
            sortedlist = characterlist
            sort_list(sort_key, sortedlist)
            reload_matrix(sortedlist)
        end,
    }
    for i = 1, numcolumns do
        matrix["ALIGNMENT" .. i] = columndefs[i].alignment
        matrix["0:" .. i] = columndefs[i].title
    end
    set_sort_mode(defsort)

    return matrix
end

local function pad(d) return ("%09d%s"):format(#d, d) end

local function sortfunc(a, b)
    return tostring(a):gsub("%d+", pad) < tostring(b):gsub("%d+", pad)
end

-- returns a hex color for a name/type combination
local function get_goods_color(name, type)
    local color
    if name:match(" Ore") then
        color = TA.GUI.Color["ore"]
    elseif name:match("Mining") or name:match("Mineral Scanner") then
        color = TA.GUI.Color["mining"]
    elseif name:match("Turret") then
        color = TA.GUI.Color["turretweapon"]
    elseif TA.GUI.Color[type] then
        color = TA.GUI.Color[type]
    else
        color = TA.GUI.Color.WHITE
    end
    return color
end

-- returns formatted location string for station id
local function get_location_string(stationid)
    local station = TA.stations[stationid]
    local class = GetStationClass(stationid)
    if class == 0 then class = "Conquerable" end
    return TA.GUI.Factioncolors[TA.GUI.Systemownership[station.systemid]] ..
        ShortLocationStr(station.sectorid) .. "\127o [" ..
        FactionName[GetStationFaction(stationid)] .. " " ..
        class .. "]" .. (station.dynamic and
            (TA.GUI.Color.ORANGE .. " D\127o") or "")
end

-- returns formatted price string
local function get_price_string(credits, crystals)
    credits = tonumber(credits)
    crystals = tonumber(crystals)
    if credits then
        if is_premium then crystals = nil end
        return CreditsPlusCrystal(credits, crystals)
    end
    return ""
end

-- extracts the credits part of a formatted price string for easy comparison
local function price_string_to_number(price)
    price = price:gsub(",", ""):match("%d+")
    return tonumber(price)
end

TA.GUI.goods_cu = iup.label { title = "", expand = "HORIZONTAL", alignment = "ARIGHT" }
TA.GUI.goods_mass = iup.label { title = "", expand = "HORIZONTAL", alignment = "ARIGHT" }
TA.GUI.goods_list_typeids = {}
TA.GUI.goods_list = iup.stationsublist { dropdown = "YES", expand = "HORIZONTAL",
    action = function(_, _, index, state)
        if state == 0 then return end
        local typeid = TA.GUI.goods_list_typeids[index]
        TA.GUI.goods_cu.title = comma_value(TA.goods[typeid].vol)
        TA.GUI.goods_mass.title = comma_value(TA.goods[typeid].mass)
        TA.GUI.goods_tbl = {}
        for k, v in pairs(TA.goods[typeid]) do
            if type(v) == "table" then
                table.insert(TA.GUI.goods_tbl, 1, {
                    name = get_location_string(k),
                    sell = get_price_string(v.s, v.s2),
                    buy = get_price_string(v.b, v.v2),
                }
                )
            end
        end
        TA.GUI.goods_matrix:reset(TA.GUI.goods_tbl)
    end
}

local goods_sorttbl = {
    [1] = {
        title = "Name",
        alignment = "ALEFT",
        fn = function(a, b) return a.name < b.name end,
    },
    [2] = {
        title = "Sell",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.buy == b.buy then return a.name < b.name
            else return (price_string_to_number(a.buy) or 0) >
                    (price_string_to_number(b.buy) or 0)
            end
        end,
    },
    [3] = {
        title = "Buy",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.sell == b.sell then return a.name < b.name
            else return (price_string_to_number(a.sell) or 9999999999) <
                    (price_string_to_number(b.sell) or 9999999999)
            end
        end,
    },
    update_entry = function(self, index, data)
        self:setcell(index, 1, " " .. data.name)
        self:setcell(index, 2, " " .. data.buy)
        self:setcell(index, 3, " " .. data.sell)
    end,
    on_sel = function(_, row, data)
        if clicked[row] then
            local location_str = filter_colorcodes(data.name):match("(.-)%[")
            if location_str then
                local sectorid = SectorIDFromLocationStr(location_str)
                if sectorid and not IsTrainingSector(sectorid) then
                    NavRoute.SetFinalDestination(sectorid)
                    TA.Print("Navigation destination set to " .. location_str)
                end
            end
            clicked = {}
            click_timer:Kill()
        else
            clicked = {}
            clicked[row] = true
            click_timer:SetTimeout(500, click_timeout)
        end
    end,
}
TA.GUI.goods_matrix = create_matrix(goods_sorttbl, 1)
local goods_tab = iup.vbox {
    iup.hbox {
        iup.fill { size = 4 },
        iup.label { title = "Item: " }, TA.GUI.goods_list,
        iup.fill { size = 5 },
        iup.label { title = "Volume: " }, TA.GUI.goods_cu,
        iup.label { title = " cu", expand = "HORIZONTAL", alignment = "ALEFT" },
        iup.fill { size = 8 },
        iup.label { title = "Mass: " }, TA.GUI.goods_mass,
        iup.label { title = " kg", expand = "HORIZONTAL", alignment = "ALEFT" },
        iup.fill { size = 4 },
        alignment = "ACENTER",
    },
    iup.fill { size = is_mobile and 6 or 0 },
    TA.GUI.goods_matrix,
    tabtitle = "Trade Goods",
}

TA.GUI.station_buy = iup.stationsublist {
    dropdown = "YES",
    action = function()
        TA.GUI.stations_tbl = {}
        local buy_id = sorted.stations[tonumber(TA.GUI.station_buy.value)][2]
        local sell_id = sorted.stations[tonumber(TA.GUI.station_sell.value)][2]
        for _, v in pairs(TA.goods) do
            if (v[buy_id] and v[buy_id].s) and
                (v[sell_id] and v[sell_id].b) and
                (v[sell_id].b > v[buy_id].s) and not
                (not is_premium and v[buy_id].s2 > (v[buy_id].b2 or 0)) then
                table.insert(TA.GUI.stations_tbl, 1, {
                    name = get_goods_color(v.name, v.type) .. v.name,
                    sell = v[sell_id].b,
                    buy = v[buy_id].s,
                    profit_cu = (v[sell_id].b - v[buy_id].s) / (v.vol or 1)
                }
                )
            end
        end
        if #TA.GUI.stations_tbl == 0 then
            table.insert(TA.GUI.stations_tbl, 1, {
                name = "No results found",
                buy = "",
                sell = "",
            })
        end
        TA.GUI.station_matrix:reset(TA.GUI.stations_tbl)
    end
}
TA.GUI.station_sell = iup.stationsublist {
    dropdown = "YES",
    action = TA.GUI.station_buy.action
}

local stations_sorttbl = {
    [1] = {
        title = "Item Name",
        alignment = "ALEFT",
        fn = function(a, b) return a.name < b.name end,
    },
    [2] = {
        title = "Buy",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.buy == b.buy then return a.name < b.name
            else return sortfunc(a.buy, b.buy)
            end
        end,
    },
    [3] = {
        title = "Sell",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.sell == b.sell then return a.name < b.name
            else return not sortfunc(a.sell, b.sell)
            end
        end,
    },
    [4] = {
        title = "Profit / cu",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.profit_cu == b.profit_cu then return a.name < b.name
            else return not sortfunc(a.profit_cu, b.profit_cu)
            end
        end,
    },
    update_entry = function(matrix, index, data)
        matrix:setcell(index, 1, " " .. data.name)
        matrix:setcell(index, 2, " " .. get_price_string(data.buy))
        matrix:setcell(index, 3, " " .. get_price_string(data.sell))
        matrix:setcell(index, 4, " " .. (data.profit_cu and format("%.2fc", data.profit_cu) or ""))
    end,
}

TA.GUI.station_matrix = create_matrix(stations_sorttbl, 4)

local function selectCurrentStation()
    local currentStation = GetCurrentStationID()
    if not currentStation then
        local sectorid = GetCurrentSectorid()
        for stationid, stationdata in pairs(TA.stations) do
            if stationdata.sectorid == sectorid then
                currentStation = stationid
                break
            end
        end
    end
    if currentStation then
        for i, stationData in pairs(sorted.stations) do
            if currentStation == stationData[2] then
                TA.GUI.station_buy.value = i
                TA.GUI.station_sell.value = i
                break
            end
        end
    end
end

local function doSelectDestinationStation(destinationSector)
    for i, stationData in pairs(sorted.stations) do
        if destinationSector == TA.stations[stationData[2]].sectorid
        then
            TA.GUI.station_sell.value = i
            return true
        end
    end
    return false
end

local function selectDestinationStation()
    local route = NavRoute.GetCurrentRoute()
    route.n = #route
    if route.n > 0 then
        local sectorid = GetCurrentSectorid()
        for i = 1, route.n do
            if route[i] ~= sectorid and doSelectDestinationStation(route[i]) then return end
        end
    end
end

local stations_tab = iup.vbox {
    iup.hbox {
        iup.fill { size = 4 },
        iup.label { title = "Buy: " }, TA.GUI.station_buy,
        iup.fill { size = is_mobile and 24 or 12 },
        iup.button { title = "", size = Font.Default .. "x" .. Font.Default,
            image = plugin_path .. "arrows.png",
            bgcolor = "255 255 255 255 *",
            action = function()
                local tmp = TA.GUI.station_buy.value
                TA.GUI.station_buy.value = TA.GUI.station_sell.value
                TA.GUI.station_sell.value = tmp
                TA.GUI.station_buy:action()
            end },
        iup.fill { size = is_mobile and 24 or 12 },
        iup.label { title = "Sell: " }, TA.GUI.station_sell,
        alignment = "ACENTER",
    },
    iup.fill { size = is_mobile and 6 or 0 },
    TA.GUI.station_matrix,
    tabtitle = "Stations",
    OnShow = function()
        selectDestinationStation()
        TA.GUI.station_buy:action()
    end,
}

-- empties matrixes and other data tables used for the GUI
local function clear_data()
    TA.GUI.station_buy[1] = nil
    TA.GUI.station_sell[1] = nil
    TA.GUI.goods_list[1] = nil
    TA.GUI.goods_list_typeids = {}
    sorted.stations = {}
    TA.GUI.station_matrix:reset({})
    TA.GUI.goods_matrix:reset({})
end

local function refresh_gui()
    clear_data()

    local tbl = {}
    for k, v in pairs(TA.goods) do
        table.insert(tbl, { k, get_goods_color(v.name, v.type) .. v.name })
    end
    table.sort(tbl, function(a, b) return a[2] < b[2] end)
    for i, v in ipairs(tbl) do
        TA.GUI.goods_list[i] = v[2]
        TA.GUI.goods_list_typeids[i] = v[1]
    end

    for k in pairs(TA.stations) do
        local location = get_location_string(k)
        table.insert(sorted.stations, { location, k })
    end
    table.sort(sorted.stations, function(a, b) return sortfunc(a[1], b[1]) end)

    for i, v in ipairs(sorted.stations) do
        TA.GUI.station_buy[i] = v[1]
        TA.GUI.station_sell[i] = v[1]
    end

    selectCurrentStation()
    selectDestinationStation()

    -- width of the matrix columns
    local width = getwidth(TA.GUI.goods_matrix) - Font.Default
    TA.GUI.goods_matrix.width1 = width / 1.5
    TA.GUI.goods_matrix.width2 = width / 6
    TA.GUI.goods_matrix.width3 = width / 6

    TA.GUI.station_matrix.width1 = width / 2 -- item name
    TA.GUI.station_matrix.width2 = width / 6 -- buy price
    TA.GUI.station_matrix.width3 = width / 6 -- sell price
    TA.GUI.station_matrix.width4 = width / 6 -- profit / cu

    -- iup sucks, better do the dropdown sizing ourselves:
    local dropdown_width = (TA.GUI.MAIN.w - (stations_tab[1][2].w +
        stations_tab[1][7].w + stations_tab[1][5].w +
        (is_mobile and 68 or 42))) / 2
    TA.GUI.station_buy.size = dropdown_width
    TA.GUI.station_sell.size = dropdown_width
    iup.Refresh(TA.GUI.MAIN)

    -- populate both tabs
    TA.GUI.goods_list:action(nil, 1, 1)
    TA.GUI.station_buy:action()
end

local function init()
    local load = unspickle(LoadSystemNotes(notesid)) or {}
    if next(load) == nil then
        -- try to load 1.13 database if we don't have a current one
        local charid = GetCharacterID()
        if charid then
            load = unspickle(LoadSystemNotes(charid .. "201")) or {}
            if next(load) ~= nil then
                TA.Print("Importing old format database")
            end
        end
    else
        TA.Print("Loading database")
    end
    TA.stations = load.stations or {}
    if load.version then
        TA.goods = load.goods or {}
        TA.goods_old = load.goods_old or {}
        TA.target_print = load.target_print or false
        toggle_target_print(TA.target_print)
        fullscreen = load.fullscreen
    else -- old format db
        TA.goods = {}
        TA.goods_old = load.goods or {}
        for _, station in pairs(TA.stations) do
            station.name = nil
        end
        save_data_now()
    end
    TA.GUI.MAIN = iup.dialog {
        iup.pdasubframebg {
            iup.vbox {
                iup.pda_root_tabs {
                    goods_tab,
                    stations_tab,
                    tabchange_cb = function(_, newtab)
                        if newtab.OnShow then newtab:OnShow() end
                    end,
                },
                TA.GUI.Close,
            },
        },
        k_any = function(_, char)
            local cmd = gkinterface.GetCommandForKeyboardBind(char)
            if cmd == command then
                TA.GUI.Hide()
                return iup.IGNORE
            end
            return iup.CONTINUE
        end,
        topmost = "YES",
        size = tostring(44 * Font.Default) .. "x%90",
        fullscreen = fullscreen and "YES" or "NO",
        defaultesc = TA.GUI.Close,
        title = TA.Title,
    }
    TA.GUI.MAIN:map()
    has_old_goods = next(TA.goods_old) ~= nil
    is_premium = IsPremiumSubscription()
    init_done = true
end

function TA.GUI.Hide()
    clear_data()
    TA.GUI.MAIN:hide()
end

function TA.GUI.Show()
    if not init_done then init() end
    refresh_gui()
    if HUD.visible then
        -- when the HUD is shown, show as a popup to prevent clicking on
        -- the HUD and bringing it to the front
        TA.GUI.MAIN:popup(iup.CENTER, iup.CENTER)
    else
        TA.GUI.MAIN:showxy(iup.CENTER, iup.CENTER)
    end
end

TA.GUI.Close = iup.stationbutton {
    title = "Close",
    expand = "HORIZONTAL",
    action = TA.GUI.Hide
}

--- new database format conversion function
local function convert_database_item(name, typeid)
    if has_old_goods then
        for stationid in pairs(TA.stations) do
            if TA.goods_old[name] and TA.goods_old[name][stationid] then
                if not TA.goods[typeid][stationid] then
                    TA.goods[typeid][stationid] = {
                        b = TA.goods_old[name][stationid].buy or nil,
                        s = TA.goods_old[name][stationid].sell or nil
                    }
                end
                TA.goods_old[name][stationid] = nil
                local has_station = false
                for _, v in pairs(TA.goods_old[name]) do
                    if type(v) == "table" then has_station = true end
                end
                if not has_station then TA.goods_old[name] = nil end
            end
        end
        has_old_goods = next(TA.goods_old) ~= nil
    end
end

-- updates station buy price in database and saves it
local function update_price(typeid, stationid, buy)
    if TA.goods[typeid] and TA.stations[stationid] then
        TA.goods[typeid][stationid] = TA.goods[typeid][stationid] or {}
        TA.goods[typeid][stationid].b = buy
    end
    save_data()
end

-- wrapper for GetBestPriceInfoOfItem(),
-- returns lprice, lstationid, gprice, gstationid or nil
local function getBestPriceOfItem(itemid)
    local lprice, lloc, gprice, gloc = GetBestPriceInfoOfItem(itemid)
    if not (lprice and lloc and gprice and gloc) then return end
    local lsectorid = tonumber(sub(lloc, 1, -3))
    local lstationid = lsectorid * 256 + tonumber(sub(lloc, -2))
    local gsectorid = tonumber(sub(gloc, 1, -3))
    local gstationid = gsectorid * 256 + tonumber(sub(gloc, -2))
    return lprice, lstationid, gprice, gstationid
end

-- updates and saves system and global best price for supplied item
local function updateBestPricesForItem(itemid, typeid, currentStationId)
    currentStationId = currentStationId or GetCurrentStationID()
    if itemid and typeid and itemid ~= typeid then
        local lprice, lstationid, gprice, gstationid = getBestPriceOfItem(itemid)
        local qty = GetInventoryItemQuantity(itemid)
        -- XXX: Any chance of splitting a single unit item off to get the unit
        -- price for consistency with merch prices?
        if lprice and lstationid and (lstationid ~= currentStationId) then
            lprice = floor(lprice / qty)
            update_price(typeid, lstationid, lprice)
        end
        if gprice and gstationid and (gstationid ~= currentStationId) then
            gprice = floor(gprice / qty)
            update_price(typeid, gstationid, gprice)
        end
    end
end

-- c/cu price coloring
local function get_price_color(price, vol)
    local profit_cu = price / vol
    if profit_cu > 7500 then return TA.GUI.Color.SPRINGGREEN end
    if profit_cu > 5000 then return TA.GUI.Color.GREEN end
    if profit_cu > 3500 then return TA.GUI.Color.YELLOW end
    if profit_cu > 2000 then return TA.GUI.Color.ORANGE end
    return TA.GUI.Color.RED
end

-- output function for target_print
local function print_item_price(name, price, count, vol)
    local color = get_price_color(price, vol)
    local function colored_string(string)
        return color .. string .. "\127o"
    end

    count = tonumber(count) or 1
    local volume_string = ", " .. (vol * count) .. "cu"
    return HUD:PrintSecondaryMsg(TA.print_prefix .. name .. " sell" ..
        (name:sub(-1) == "s" and "" or "s") .. " for " ..
        colored_string(get_price_string(price)) ..
        (count > 1 and " (all " .. count .. " for " ..
            colored_string(get_price_string(count * price)) ..
            volume_string .. ")" or volume_string)
    )
end

-- target (cargo) sell price and cu info
local function target_print(name, count, stationid, systemid)
    for _, data in pairs(TA.goods) do
        if data.name == name then
            -- check current sector
            if stationid and data[stationid] and data[stationid].b then
                return print_item_price(name, data[stationid].b, count, data.vol)
            end
            -- nothing in sector, check best price in system
            local bestprice = -1
            for _stationid, stationdata in pairs(TA.stations) do
                if stationdata.systemid == systemid and
                    data[_stationid] and data[_stationid].b
                then
                    local stationprice = data[_stationid].b
                    if stationprice and stationprice > bestprice then
                        bestprice = stationprice
                    end
                end
            end
            if bestprice > 0 then
                return print_item_price(name, bestprice, count, data.vol)
            end
            -- nothing in system, check overall best
            for _, itemdata in pairs(data) do
                if type(itemdata) == "table" and itemdata.b and
                    itemdata.b > bestprice
                then
                    bestprice = itemdata.b
                end
            end
            if bestprice >= 0 then
                return print_item_price(name, bestprice, count, data.vol)
            end
            -- looks like we don't have any price for this one
            -- (we could look for buy prices but we won't)
            return HUD:PrintSecondaryMsg(TA.print_prefix .. name ..
                " has no known sell price")
        end
        -- item not in database
    end
end

-- returns stationid if current sector has a known station
local function get_station_in_current_sector()
    local sectorid = GetCurrentSectorid()
    for stationid, stationdata in pairs(TA.stations) do
        if stationdata.sectorId == sectorid then
            return stationid
        end
    end
end

function TA.EventHandler(event, ...)
    if not init_done then init() end
    if event == "SYSTEM_BEST_PRICE" then
        local args = { ... } -- itemid, price, locationid
        if not args[2] then return end -- no system best price in unknown system
        local itemid = args[1]
        local typeid = GetInventoryItemType(itemid)
        if TA.goods[typeid] then -- can only update if we have seen the item
            -- now we don't actually use SYSTEM_BEST_PRICE directly because it
            -- does not provide the galactic best price:
            updateBestPricesForItem(itemid, typeid)
        end
    elseif event == "TARGET_CHANGED" and TA.target_print then
        local systemid = GetCurrentSystemid()
        local target_name = GetTargetInfo() or ""
        if last_scanned ~= target_name then last_scanned = "" end
        local name, count = target_name:match("(.-) %((%d+x)")
        if not name then name = target_name:match("(.-) %(") end
        count = count and count:sub(1, -2) or 1
        if not name then return end
        local stationid = get_station_in_current_sector()
        target_print(name, count, stationid, systemid)
    elseif event == "TARGET_SCANNED" and TA.target_print then
        local target_name = GetTargetInfo() or ""
        if last_scanned == target_name then return end
        local systemid = GetCurrentSystemid()
        local stationid = get_station_in_current_sector()
        local args = { ... }
        if args[1] ~= "Ship too far to scan\n" then
            last_scanned = target_name
        end
        for line in args[1]:gmatch("[^\r\n]+") do
            for name, count in line:gmatch("(.-) x(%d+)") do
                target_print(name, count, stationid, systemid)
            end
        end
    elseif (event == "ENTERED_STATION" or event == "STATION_UPDATE_PRICE") and
        not GetStationType() and GetCurrentStationID()
    then
        if shipMassBugWorkaround then activeShipMass = GetActiveShipMass() end
        local stationid = GetCurrentStationID()

        -- add station, if new
        if not TA.stations[stationid] then
            TA.Print("Adding current station to database.")
            TA.stations[stationid] = {
                sectorid = GetCurrentSectorid(),
                systemid = GetCurrentSystemid()
            }
        end
        -- update station dynamic economy status
        TA.stations[stationid].dynamic = IsDynamicEconomyStation()

        -- add/update items sold by the station
        local station_merch = {}
        for i = 1, GetNumStationMerch() do
            local info = GetStationMerchInfo(i)
            if info and info.type then
                station_merch[info.itemtype] = true
                TA.goods[info.itemtype] = TA.goods[info.itemtype] or {}
                TA.goods[info.itemtype][stationid] = TA.goods[info.itemtype][stationid] or {}
                TA.goods[info.itemtype][stationid].s = info.price
                TA.goods[info.itemtype][stationid].s2 = info.price and info.price2
                TA.goods[info.itemtype].name = info.name
                TA.goods[info.itemtype].mass = floor(1000 * (info.mass -
                    (info.type == "ship" and activeShipMass or 0)) + 0.5)
                TA.goods[info.itemtype].vol = info.volume
                TA.goods[info.itemtype].type = info.type
                convert_database_item(info.name, info.itemtype)
                -- station merch is always in the player.invbestprices table,
                -- so let's get them, too:
                updateBestPricesForItem(info.itemid, info.itemtype, stationid)
            end
        end

        -- add/update items in station and ship inventory
        for _, info in StationSellableInventoryPairs() do
            -- check if we got a type, otherwise it's not sellable, e.g. ship engines
            if info.type then
                TA.goods[info.itemtype] = TA.goods[info.itemtype] or {}
                TA.goods[info.itemtype][stationid] = TA.goods[info.itemtype][stationid] or {}
                TA.goods[info.itemtype][stationid].b = info.price
                TA.goods[info.itemtype][stationid].b2 = info.price and info.price2
                TA.goods[info.itemtype].name = info.name
                TA.goods[info.itemtype].mass = floor(1000 *
                    (InvManager.GetInventoryItemPrototype(info.itemtype):GetMass() -
                        (info.type == "ship" and activeShipMass or 0)) + 0.5)
                TA.goods[info.itemtype].vol = TA.goods[info.itemtype].vol or
                    GetInventoryItemVolume(info.itemid)
                TA.goods[info.itemtype].type = info.type
                -- best price is not available for station inventory, and for
                -- ship inventory we already get via SYSTEM_BEST_PRICE above
                --updateBestPricesForItem(v.itemid, v.itemtype, stationid)
            end
        end

        -- purge old sell prices for items no longer for sale at this station
        for typeid in pairs(TA.goods) do
            if TA.goods[typeid][stationid] and not station_merch[typeid] then
                if TA.goods[typeid][stationid].b then
                    TA.goods[typeid][stationid].s = nil
                    TA.goods[typeid][stationid].s2 = nil
                else
                    TA.goods[typeid][stationid] = nil
                end
            end
        end

        if event == "ENTERED_STATION" then
            -- don't delay saving after docking
            save_data_now()
        else
            save_data()
        end
    elseif event == "PLAYER_LOGGED_OUT" or event == "UNLOAD_INTERFACE" then
        if timer:IsActive() then
            timer:Kill()
            save_data_now()
        end
    end
end

function TA.CommandHandler(_, args)
    if args then
        local cmd = args[1]:lower()
        if cmd == "version" then
            TA.Print(TA.Title)
        elseif cmd == "reset" then
            if not init_done then init() end
            -- TODO: This should have a confirmation
            TA.goods = {}
            TA.goods_old = {}
            TA.stations = {}
            save_data()
            refresh_gui()
            TA.Print("Database has been reset.")
        elseif cmd == "target_print" then
            if not init_done then init() end
            toggle_target_print()
            save_data_now()
        elseif cmd == "fullscreen" then
            if not init_done then init() end
            fullscreen = not fullscreen
            TA.Print("GUI switched to " ..
                (fullscreen and "fullscreen" or "windowed") ..
                " mode, will reload the interface now.")
            save_data_now()
            ReloadInterface()
        end
    else
        TA.GUI.Show()
    end
end

RegisterUserCommand(command, TA.CommandHandler)
RegisterEvent(TA.EventHandler, "SYSTEM_BEST_PRICE")
RegisterEvent(TA.EventHandler, "ENTERED_STATION")
RegisterEvent(TA.EventHandler, "STATION_UPDATE_PRICE")
RegisterEvent(TA.EventHandler, "PLAYER_LOGGED_OUT")
RegisterEvent(TA.EventHandler, "UNLOAD_INTERFACE")
RegisterEvent(TA.EventHandler, "TARGET_CHANGED")
RegisterEvent(TA.EventHandler, "TARGET_SCANNED")
